/*******************************************************************************
 * Copyright 2007  Hon Hai Precision Ind. Co. Ltd.
 * All Rights Reserved.
 * No portions of this material shall be reproduced in any form without the
 * written permission of Hon Hai Precision Ind. Co. Ltd.
 *
 * All information contained in this document is Hon Hai Precision Ind.
 * Co. Ltd. company private, proprietary, and trade secret property and
 * are protected by international intellectual property laws and treaties.
 *
 ******************************************************************************/

/*******************************************************************************
 ***  Filename:
 ***      swresetd.c
 ***
 ***  Description:
 ***      Check reset button
 ***
 ***  History:
 ***
 ***   Modify Reason                    Author          Date            Search Flag(Option)
 ***--------------------------------------------------------------------------------------
 ***   Check Reset Button               pling           2006/04/17
 ***   Port to U12H081 platform         pling           2006/10/22
 ***   Port to U12H093, add WPS func    pling           2007/8/16
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#include <epivers.h>
#include <shutils.h>
#include <bcmdevs.h>
#include <bcmparams.h>
#include <bcmgpio.h>
#include <bcmnvram.h>

#include "wps_monitor.h"


#define GPIO_RESET_BUTTON               6
#define GPIO_POWER_GREEN                5

#define GPIO_RESET_BUTTON_MASK          1 << GPIO_RESET_BUTTON

#define ACTION_NONE                     0
#define ACTION_REBOOT                   1
#define ACTION_LOAD_DEFAULT             2

#define TMP_UPGRADE_FILE                "/tmp/upgrade.tmp"

int check_upgrade_status(void)
{
    int fd;

    /* Check tmp file */
    if ((fd = open(TMP_UPGRADE_FILE, O_RDONLY)) < 0)
        return 0;   /* No firmware upgrade in progress */
    
    return 1;       /* upgrade in progress */

}

static int check_reset_button(int gpio_status)
{
    unsigned long value;
    int is_loaddefault = 1;
    int i;
    
    /* Reset button is active low */
    value = gpio_status & GPIO_RESET_BUTTON_MASK;
    if (value)
        return ACTION_NONE;

#if (!defined FW_TYPE_JP && !defined FW_TYPE_KO )
    /* Change Power LED to amber, when reset button is first pressed */
    bcmgpio_out(1 << GPIO_POWER_GREEN, 0);
#endif

    /* Wait up to 5 second for user load default */
    for (i=0; i<50; i++)
    {
        bcmgpio_in(1 << GPIO_RESET_BUTTON, &value);
        if (value)
        {
            is_loaddefault = 0;
            break;
        }
        usleep(100000);
    }
    
    /* Reboot */
    if (!is_loaddefault)
    {
        fprintf(stderr, "Reset\n");
        return ACTION_REBOOT;
    }
 
    /*
     * User has not release reset button yet, 
     * Blink Power LED every sec, 0.25 sec on, 0.75 sec off
     */    
    /* 5 sec passed, need to load default */
    while (1)
    {
        bcmgpio_in(1 << GPIO_RESET_BUTTON, &value);
        if (value)
            break;

        bcmgpio_out(1 << GPIO_POWER_GREEN, 0);
        usleep(750000);

        bcmgpio_in(1 << GPIO_RESET_BUTTON, &value);
        if (value)
            break;
        
        bcmgpio_out(1 << GPIO_POWER_GREEN, 1 << GPIO_POWER_GREEN);
        usleep(250000);
    }

    fprintf(stderr, "Load Default\n");

    /* Turn off power LED */
    bcmgpio_out(1 << GPIO_POWER_GREEN, 0);

    return ACTION_LOAD_DEFAULT;
}

static int do_reboot(int action)
{
    if (action == ACTION_LOAD_DEFAULT)
    {
        /* Pull low Ethernet PHY according to Netgear Ken's requirement */
        system("et robowr 0x0 0xF 0x0F"); /*setup the POWRN DOWN MODE register*/

        system("/usr/sbin/nvram loaddefault");
    }
    
    system("/sbin/reboot");

    return 0;
}

/* Do wireless band change for BCM4321 */
#if (defined FW_TYPE_JP || defined FW_TYPE_KO) 
static int do_wlan_band_change(void)
{
    if (nvram_match("wla_band", "5G"))
    {
        nvram_set("wla_band", "2.4G");
        nvram_set("wla_channel", "3");
        nvram_set("wlg_wlanstate", "Disable");
    }
    else
    {
        nvram_set("wla_band", "5G");
        nvram_set("wla_channel", "36");
        nvram_set("wlg_wlanstate", "Enable");
    }
    nvram_set("wl_mode_changed", "1");
    nvram_commit();

    eval("rc", "wlanrestart");
    return 0;
}
#endif

int main(int argc, char *argv[])
{
    int status;
    int gpio_status;
    
    /* Daemonize myself */
    if (daemon (1, 1) == -1)
        perror ("daemon");

    /* Setup GPIOs for swresetd */
    bcmgpio_connect(GPIO_POWER_GREEN, BCMGPIO_DIRN_OUT);
    bcmgpio_connect(GPIO_RESET_BUTTON, BCMGPIO_DIRN_IN);

    /* Setup GPIOs for WPS function */
    bcmgpio_connect(GPIO_WPS_BUTTON,  BCMGPIO_DIRN_IN);
    bcmgpio_connect(GPIO_11N_ANT0_TX, BCMGPIO_DIRN_IN);
    bcmgpio_connect(GPIO_11N_ANT1_TX, BCMGPIO_DIRN_IN);
    bcmgpio_connect(GPIO_11G_ANT,     BCMGPIO_DIRN_IN);

    while (1)
    {
        /* 
         * Read all GPIOs of interest at the same time
         * to reduce # of IOCTLs
         */
        bcmgpio_in(GPIO_RESET_BUTTON_MASK |
                   GPIO_WPS_BUTTON_MASK   |
                   GPIO_11N_ANT0_TX_MASK  |
                   GPIO_11N_ANT1_TX_MASK  |
                   GPIO_11G_ANT_MASK,
                   &gpio_status);

        status = check_reset_button(gpio_status);

        if (status == ACTION_REBOOT)
        {
            /* Make sure we are not doing firmware upgrade */
            if (check_upgrade_status())
                printf("Upgrading firmware, don't reboot\n");
            else
            {
                /* For JP firmware, this button does not reset. 
                 * Instead it changes BCM4321 between 2.4GHz <-> 5GHz.
                 */
#if (defined FW_TYPE_JP || defined FW_TYPE_KO )
                do_wlan_band_change();
#else
                do_reboot(ACTION_REBOOT);
#endif
            }
        }
        else
        if (status  == ACTION_LOAD_DEFAULT)
        {
            /* 5 sec passed, load default */
            /* Make sure we are not doing firmware upgrade */
            if (check_upgrade_status())
                printf("Upgrading firmware, don't load default\n");
            else
                do_reboot(ACTION_LOAD_DEFAULT);
        }

        /* Do WPS check */
        wps_check(gpio_status);

        usleep(200000);     /* 2007/12/17: per NETGEAR Matt request, update WLAN LED every 200ms */
    }
}
